package id.walt.dif

import id.walt.model.dif.PresentationDefinition
import id.walt.model.oidc.klaxon
import io.kotest.core.spec.style.AnnotationSpec
import io.kotest.matchers.shouldBe
import io.kotest.matchers.shouldNotBe

class DifTest: AnnotationSpec() {

  @Test
  fun testParsePresentationDefinition() {
    val pd = """{
            "id": "vp token example",
            "input_descriptors": [
                {
                    "id": "id card credential",
                    "format": {
                        "ldp_vc": {
                            "proof_type": [
                                "Ed25519Signature2018"
                            ]
                        }
                    },
                    "constraints": {
                        "fields": [
                            {
                                "path": [
                                    "${'$'}.type"
                                ],
                                "filter": {
                                    "type": "string",
                                    "pattern": "IDCardCredential"
                                }
                            }
                        ]
                    }
                }
            ]
        }"""

    val parsedPD = klaxon.parse<PresentationDefinition>(pd)
    parsedPD shouldNotBe null
    parsedPD!!.id shouldBe "vp token example"
    parsedPD!!.input_descriptors shouldNotBe null
    parsedPD!!.input_descriptors!!.size shouldBe 1
    parsedPD!!.input_descriptors[0].id shouldBe "id card credential"
    parsedPD!!.input_descriptors[0].format shouldNotBe null
  }

  @Test
  fun testParsePresentationDefinitionWithSubmissionRequirements() {
    val pd = """{
    "id": "32f54163-7166-48f1-93d8-ff217bdb0653",
    "submission_requirements": [
      {
        "name": "Banking Information",
        "purpose": "We can only remit payment to a currently-valid bank account in the US, Germany or France.",
        "rule": "pick",
        "count": 1,
        "from": "A"
      },
      {
        "name": "Employment Information",
        "purpose": "We are only verifying one current employment relationship, not any other information about employment.",
        "rule": "all",
        "from": "B"
      },
      {
        "name": "Eligibility to Drive on US Roads",
        "purpose": "We need to verify eligibility to drive on US roads via US or EU driver's license, but no biometric or identifying information contained there.",
        "rule": "pick",
        "count": 1,
        "from": "C"
      }
    ],
    "input_descriptors": [
      {
        "id": "banking_input_1",
        "name": "Bank Account Information",
        "purpose": "Bank Account required to remit payment.",
        "group": ["A"],
        "constraints": {
          "limit_disclosure": "required",
          "fields": [
            {
              "path": ["${'$'}.credentialSchema", "${'$'}.vc.credentialSchema"],
              "filter": {
                "allOf": [
                  {
                    "type": "array",
                    "contains": {
                      "type": "object",
                      "properties": {
                        "id": {
                          "type": "string",
                          "pattern": "https://bank-standards.example.com#accounts"
                        }
                      },
                      "required": ["id"]
                    }
                  },
                  {
                    "type": "array",
                    "contains": {
                      "type": "object",
                      "properties": {
                        "id": {
                          "type": "string",
                          "pattern": "https://bank-standards.example.com#investments"
                        }
                      },
                      "required": ["id"]
                    }
                  }
                ]
              }
            },
            {
              "path": ["${'$'}.issuer", "${'$'}.vc.issuer", "${'$'}.iss"],
              "purpose": "We can only verify bank accounts if they are attested by a trusted bank, auditor or regulatory authority.",
              "filter": {
                "type": "string",
                "pattern": "did:example:123|did:example:456"
              }
            },
            {
              "path": [
                "${'$'}.credentialSubject.account[*].account_number",
                "${'$'}.vc.credentialSubject.account[*].account_number",
                "${'$'}.account[*].account_number"
              ],
              "purpose": "We can only remit payment to a currently-valid bank account in the US, France, or Germany, submitted as an ABA Acct # or IBAN.",
              "filter": {
                "type": "string",
                "pattern": "^[0-9]{10-12}|^(DE|FR)[0-9]{2}\\s?([0-9a-zA-Z]{4}\\s?){4}[0-9a-zA-Z]{2}${'$'}"
              }
            },
            {
              "path": ["${'$'}.credentialSubject.portfolio_value", "${'$'}.vc.credentialSubject.portfolio_value", "${'$'}.portfolio_value"],
              "purpose": "A current portfolio value of at least one million dollars is required to insure your application",
              "filter": {
                "type": "number",
                "minimum": 1000000
              }
            }
          ]
        }
      },
      {
        "id": "banking_input_2",
        "name": "Bank Account Information",
        "purpose": "We can only remit payment to a currently-valid bank account.",
        "group": ["A"],
        "constraints": {
          "fields": [
            {
              "path": ["${'$'}.credentialSchema.id", "${'$'}.vc.credentialSchema.id"],
              "filter": {
                "type": "string",
                "pattern": "https://bank-schemas.org/1.0.0/accounts.json|https://bank-schemas.org/2.0.0/accounts.json"
              }
            },
            {
              "path": [
                "${'$'}.issuer",
                "${'$'}.vc.issuer",
                "${'$'}.iss"
              ],
              "purpose": "We can only verify bank accounts if they are attested by a trusted bank, auditor or regulatory authority.",
              "filter": {
                "type": "string",
                "pattern": "did:example:123|did:example:456"
              }
            },
            {
              "path": [
                "${'$'}.credentialSubject.account[*].id",
                "${'$'}.vc.credentialSubject.account[*].id",
                "${'$'}.account[*].id"
              ],
              "purpose": "We can only remit payment to a currently-valid bank account in the US, France, or Germany, submitted as an ABA Acct # or IBAN.",
              "filter": {
                "type": "string",
                "pattern": "^[0-9]{10-12}|^(DE|FR)[0-9]{2}\\s?([0-9a-zA-Z]{4}\\s?){4}[0-9a-zA-Z]{2}${'$'}"
              }
            },
            {
              "path": [
                "${'$'}.credentialSubject.account[*].route",
                "${'$'}.vc.credentialSubject.account[*].route",
                "${'$'}.account[*].route"
              ],
              "purpose": "We can only remit payment to a currently-valid account at a US, Japanese, or German federally-accredited bank, submitted as an ABA RTN or SWIFT code.",
              "filter": {
                "type": "string",
                "pattern": "^[0-9]{9}|^([a-zA-Z]){4}([a-zA-Z]){2}([0-9a-zA-Z]){2}([0-9a-zA-Z]{3})?${'$'}"
              }
            }
          ]
        }
      },
      {
        "id": "employment_input",
        "name": "Employment History",
        "purpose": "We are only verifying one current employment relationship, not any other information about employment.",
        "group": ["B"],
        "constraints": {
          "limit_disclosure": "required",
          "fields": [
            {
              "path": ["${'$'}.credentialSchema", "${'$'}.vc.credentialSchema"],
              "filter": {
                "type": "string",
                "const": "https://business-standards.org/schemas/employment-history.json"
              }
            },
            {
              "path": ["${'$'}.jobs[*].active"],
              "filter": {
                "type": "boolean",
                "pattern": "true"
              }
            }
          ]
        }
      },
      {
        "id": "drivers_license_input_1",
        "name": "EU Driver's License",
        "group": ["C"],
        "constraints": {
          "fields": [
            {
              "path": ["${'$'}.credentialSchema.id", "${'$'}.vc.credentialSchema.id"],
              "filter": {
                "type": "string",
                "const": "https://schema.eu/claims/DriversLicense.json"
              }
            },
            {
              "path": ["${'$'}.issuer", "${'$'}.vc.issuer", "${'$'}.iss"],
              "purpose": "We can only accept digital driver's licenses issued by national authorities of EU member states or trusted notarial auditors.",
              "filter": {
                "type": "string",
                "pattern": "did:example:gov1|did:example:gov2"
              }
            },
            {
              "path": ["${'$'}.credentialSubject.dob", "${'$'}.vc.credentialSubject.dob", "${'$'}.dob"],
              "purpose": "We must confirm that the driver was at least 21 years old on April 16, 2020.",
              "filter": {
                "type": "string",
                "format": "date",
                "formatMaximum": "1999-05-16"
              }
            }
          ]
        }
      },
      {
        "id": "drivers_license_input_2",
        "name": "Driver's License from one of 50 US States",
        "group": ["C"],
        "constraints": {
          "fields": [
            {
              "path": ["${'$'}.credentialSchema.id", "${'$'}.vc.credentialSchema.id"],
              "filter": {
                "type": "string",
                "const": "hub://did:foo:123/Collections/schema.us.gov/american_drivers_license.json"
              }
            },
            {
              "path": ["${'$'}.issuer", "${'$'}.vc.issuer", "${'$'}.iss"],
              "purpose": "We can only accept digital driver's licenses issued by the 50 US states' automative affairs agencies.",
              "filter": {
                "type": "string",
                "pattern": "did:example:gov1|did:web:dmv.ca.gov|did:example:oregonDMV"
              }
            },
            {
              "path": ["${'$'}.credentialSubject.birth_date", "${'$'}.vc.credentialSubject.birth_date", "${'$'}.birth_date"],
              "purpose": "We must confirm that the driver was at least 21 years old on April 16, 2020.",
              "filter": {
                "type": "string",
                "format": "date",
                "forrmatMaximum": "1999-05-16"
              }
            }
          ]
        }
      }
    ]
  }"""

    val parsedPD = klaxon.parse<PresentationDefinition>(pd)
    parsedPD shouldNotBe null
  }
}
